//
// Breakout game
//
// (C)2000
// Brian Postma
// b.postma@hetnet.nl
//

#include "Breakout.h"

Breakout::Breakout()
{
    int col=16;
    int colordelta=255/(numlines-1);
    int j;

    GpSetPaletteEntry(0,  16, 24, 64);
    GpSetPaletteEntry(1,  255,255,255);
    GpSetPaletteEntry(2,   96,128,255);
    GpSetPaletteEntry(3,   255,160,64);

    bricksperline=(GetScreenWidth()-2*borderwidth)/(brickwidth+brickspace);
    width=bricksperline*(brickwidth+brickspace)+(2*borderwidth);
    showbrick=new bool[bricksperline*numlines];
    ingame=false;
    balldx=0;
    batdpos=0;
    balldy=0;

    for (j=0; j<numlines; j++)
    {
      GpSetPaletteEntry(col+j, 255,j*colordelta,255-j*colordelta);
    }
    GameInit();
}


Breakout::~Breakout()
{
}


void Breakout::GameInit()
{
    batpos=(width-batwidth)/2;
    ballx=(width-ballsize)/2;
    bally=(GetScreenHeight()-ballsize-scoreheight-2*borderwidth);
    player1score=0;
    ballsleft=3;
    dxval=2;
    if (Random()<0.5)
      balldx=dxval;
    else
      balldx=-dxval;
    balldy=-dxval;
    batdpos=0;
    InitBricks();
}


void Breakout::InitBricks()
{
    int i;
    for (i=0; i<numlines*bricksperline; i++)
      showbrick[i]=true;
}


void Breakout::KeyDown(int key)
{
    if (ingame)
    {
      if (key == GPC_VK_LEFT)
          batdpos=-3;
      if (key == GPC_VK_RIGHT)
        batdpos=3;
      if (key == GPC_VK_SELECT)
        ingame=false;
    }
    else
    {
      if (key == GPC_VK_START)
      {
        ingame=true;
        GameInit();
      }
    }
}


void Breakout::KeyUp(int key)
{
    if (key == GPC_VK_LEFT || key == GPC_VK_RIGHT)
       batdpos=0;
}


void Breakout::Paint()
{
    ClrScreen();
    if (ingame)
      PlayGame();
    else
      ShowIntroScreen();
}


void Breakout::PlayGame()
{
    MoveBall();
    CheckBat();
    CheckBricks();
    DrawPlayField();
    DrawBricks();
    ShowScore();
}


void Breakout::ShowIntroScreen()
{
    char s1[]="BREAKOUT";
    char s2[]="(c)2000,2003 by Brian Postma";
    char s3[]="B.Postma@HetNet.nl";
    char s4[]="http://www.brianpostma.com";

    MoveBall();
    CheckBat();
    CheckBricks();
    BatDummyMove();
    DrawPlayField();
    DrawBricks();
    ShowScore();
    
    DrawString(s1,(GetScreenWidth()-8*strlen(s1))/2,80,2);
    DrawString(s2,(GetScreenWidth()-8*strlen(s2))/2,100,1);
    DrawString(s3,(GetScreenWidth()-8*strlen(s3))/2,120,3);
    DrawString(s4,(GetScreenWidth()-8*strlen(s4))/2,135,3);
}


void Breakout::DrawBricks()
{
    int i,j,col=16;
    bool nobricks=true;

    for (j=0; j<numlines; j++)
    {
      for (i=0; i<bricksperline; i++)
      {
        if (showbrick[j*bricksperline+i])
        {
          nobricks=false;
          FillRect(borderwidth+i*(brickwidth+brickspace), startline+j*(brickheight+brickspace),
               brickwidth, brickheight,col+j);
        }
      }
    }
    if (nobricks)
    {
      InitBricks();
      if (ingame)
        player1score+=100;
    }
}
   

void Breakout::DrawPlayField()
{
    FillRect(0,0,width,borderwidth,1);
    FillRect(0,0,borderwidth,GetScreenHeight(),1);
    FillRect(width-borderwidth,0,borderwidth,GetScreenHeight(),1);
    FillRect(batpos,GetScreenHeight()-2*borderwidth-scoreheight, batwidth,batheight,1); // bat
    FillRect(ballx,bally,ballsize,ballsize,1); // ball
}


void Breakout::ShowScore()
{
    char buf[16];

    sprintf(buf,"Score: %05d",player1score);
    DrawString(buf,40,GetScreenHeight()-16,1);
    sprintf(buf,"Balls left: %1d",ballsleft);
    DrawString(buf,width-20-strlen(buf)*10,GetScreenHeight()-16,1);
}


void Breakout::MoveBall()
{
    ballx+=balldx;
    bally+=balldy;
    if (bally<=borderwidth)
    {
      balldy=-balldy;
      bally=borderwidth;
    }
    if (bally>=(GetScreenHeight()-ballsize-scoreheight))
    {
      if (ingame)
      {
        ballsleft--;
        if (ballsleft<=0)
          ingame=false;
      }
      ballx=batpos+(batwidth-ballsize)/2;
      bally=startline+numlines*(brickheight+brickspace);
      balldy=dxval;
      balldx=0;
    }
    if (ballx>=(width-borderwidth-ballsize))
    {
      balldx=-balldx;
      ballx=width-borderwidth-ballsize;
    }
    if (ballx<=borderwidth)
    {
      balldx=-balldx;
      ballx=borderwidth;
    }
}


void Breakout::BatDummyMove()
{
    if (ballx<(batpos+2))
      batpos-=3;
    else if (ballx>(batpos+batwidth-3))
      batpos+=3;
}


void Breakout::CheckBat()
{
    batpos+=batdpos;

    if (batpos<borderwidth)
      batpos=borderwidth;
    else if (batpos>(width-borderwidth-batwidth))
      batpos=(width-borderwidth-batwidth);
 
    if (bally>=(GetScreenHeight()-scoreheight-2*borderwidth-ballsize) && 
        bally<(GetScreenHeight()-scoreheight-2*borderwidth) &&
        (ballx+ballsize)>=batpos && ballx<=(batpos+batwidth))
    {
      bally=GetScreenHeight()-scoreheight-ballsize-borderwidth*2;
      balldy=-dxval;
      balldx=CheckBatBounce(balldx,ballx-batpos);
    }
}


int Breakout::CheckBatBounce(int dy, int delta)
{
    int sign;
    int stepsize, i=-ballsize, j=0;
    stepsize=(ballsize+batwidth)/8;

    if (dy>0)
      sign=1;
    else
      sign=-1;

    while(i<batwidth && delta>i)
    {
      i+=stepsize;
      j++;
    }
    switch(j)
    {
      case 0:
      case 1:
        return -4;
      case 2:
	return -3;
      case 7:
        return 3;
      case 3:
      case 6:
        return sign*2;
      case 4:
      case 5:
        return sign*1;
      default:
        return 4;
    }
}


void Breakout::CheckBricks()
{
    int i,j,x,y;
    int xspeed=balldx;
    if (xspeed<0) xspeed=-xspeed;
    int ydir=balldy;

    if (bally<(startline-ballsize) || bally>(startline+numlines*(brickspace+brickheight)))
      return;
    for (j=0; j<numlines; j++)
    {
      for (i=0; i<bricksperline; i++)
      {
        if (showbrick[j*bricksperline+i])
        {
          y=startline+j*(brickspace+brickheight);
          x=borderwidth+i*(brickspace+brickwidth);
          if (bally>=(y-ballsize) && bally<(y+brickheight) &&
              ballx>=(x-ballsize) && ballx<(x+brickwidth))
          {
            showbrick[j*bricksperline+i]=false;
            if (ingame)
              player1score+=(numlines-j);
            // Where did we hit the brick
            if (ballx>=(x-ballsize) && ballx<=(x-ballsize+3))
            { // leftside
              balldx=-xspeed;
            }
            else if (ballx<=(x+brickwidth-1) && ballx>=(x+brickwidth-4))
            { // rightside
              balldx=xspeed;
            }              
            balldy=-ydir;
          }
        }
      }
    }
}
